// H3DModel.cpp: implementation of the H3DModel class.
//
//////////////////////////////////////////////////////////////////////

#include "ph.h"
#include "H3DModel.h"
#include "3ds.h"
#include "HTextureStore.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

H3DModel::H3DModel()
{
	memset(&m_model, 0, sizeof(m_model));
	m_LightEnabled = false;
	m_loaded = false;
}

H3DModel::~H3DModel()
{
	int i;
	// zruseni textur
	for (i = 0; i < m_model.numOfMaterials; i++)
		htexture.RemoveTexture(m_model.pMaterials[i].textureId);
	
	// zruseni csech objektu v modelu
	for(i = 0; i < m_model.numOfObjects; i++)
	{
		// Free the faces, normals, vertices, and texture coordinates.
		delete [] m_model.pObject[i].pFaces;
		delete [] m_model.pObject[i].pNormals;
		delete [] m_model.pObject[i].pVerts;
		delete [] m_model.pObject[i].pTexVerts;
	}
}

void H3DModel::Render()
{
	if (!m_loaded) return;
	// I am going to attempt to explain what is going on below up here as not to clutter the
	// code below.  We have a model that has a certain amount of objects and textures.  We want
	// to go through each object in the model, bind it's texture map to it, then render it.
	// To render the current object, we go through all of it's faces (Polygons).
	// What is a face you ask?  A face is just (in this case) a triangle of the object.
	// For instance, a cube has 12 faces because each side has 2 triangles.
	// You might be thinking.  Well, if there are 12 faces in a cube, that makes
	// 36 vertices that we needed to read in for that object.  Not really true.  Because
	// a lot of the vertices are the same, since they share sides, they only need to save
	// 8 vertices, and ignore the duplicates.  Then, you have an array of all the
	// unique vertices in that object.  No 2 vertices will be the same.  This cuts down
	// on memory.  Then, another array is saved, which is the index numbers for each face,
	// which index in to that array of vertices.  That might sound silly, but it is better
	// than saving tons of duplicate vertices.  The same thing happens for UV coordinates.
	// You don't save duplicate UV coordinates, you just save the unique ones, then an array
	// that index's into them.  This might be confusing, but most 3D files use this format.
	// This loop below will stay the same for most file formats that you load, so all you need
	// to change is the loading code.  You don't need to change this loop (Except for animation).

	glPushMatrix();

	// zapnem/vypnem osvetleni
	if (m_LightEnabled) glEnable(GL_LIGHTING);								
	else glDisable(GL_LIGHTING);

   // Projdeme vsechny objekty a vsechny nakreslime
	for(int i = 0; i < m_model.numOfObjects; i++)
	{
		// Make sure we have valid objects just in case. (size() is in the vector class)
		if(m_model.pObject.size() <= 0) break;

		// vezmeme si pointer na aktualne kresleny objekt
		t3DObject *pObject = &m_model.pObject[i];

		// Zjistime, zda ma onbjekt texturu apripadne ji nastavime
		if (pObject->bHasTexture) 
			// Nastavime texturu podle materialu
			htexture.SetTexture(m_model.pMaterials[pObject->materialID].textureId);
		 else htexture.SetNullTexture(); // Zakazeme texturovani
			
		// Tady je mozne zmenit, zda se ma kreslit sitovy nebo plny model
		glBegin(GL_TRIANGLES);					// Begin drawing with our selected mode (triangles or lines)

			// Go through all of the faces (polygons) of the object and draw them
			for(int j = 0; j < pObject->numOfFaces; j++)
			{
				// Go through each corner of the triangle and draw it.
				for(int whichVertex = 0; whichVertex < 3; whichVertex++)
				{
					// Get the index for each point of the face
					int index = pObject->pFaces[j].vertIndex[whichVertex];

					// Give OpenGL the normal for this vertex.
					glNormal3f(pObject->pNormals[ index ].x, pObject->pNormals[ index ].y, pObject->pNormals[ index ].z);

					// If the object has a texture associated with it, give it a texture coordinate.
					if(pObject->bHasTexture) {

						// Make sure there was a UVW map applied to the object or else it won't have tex coords.
						if(pObject->pTexVerts) {
							glTexCoord2f(pObject->pTexVerts[ index ].x, pObject->pTexVerts[ index ].y);
						}
					} else {

						// Make sure there is a valid material/color assigned to this object.
						// You should always at least assign a material color to an object,
						// but just in case we want to check the size of the material list.
						// if the size is at least one, and the material ID != -1,
						// then we have a valid material.
						if(m_model.pMaterials.size() && pObject->materialID >= 0)
						{
							// Get and set the color that the object is, since it must not have a texture
							BYTE *pColor = m_model.pMaterials[pObject->materialID].color;

							// Assign the current color to this model
							glColor3ub(pColor[0], pColor[1], pColor[2]);
						}
					}

					// Pass in the current vertex of the object (Corner of current face)
					glVertex3f(pObject->pVerts[ index ].x, pObject->pVerts[ index ].y, pObject->pVerts[ index ].z);
				}
			}

		glEnd();								// End the drawing
	}
   // na konci osvetleni radsi vypnem
	glDisable(GL_LIGHTING);
	
	glPopMatrix();
}

////////////////////////////////////////////////////////////////////////////////////////

bool H3DModel::Load3DObjectTXT(const char *file)
{
	FILE *f = fopen(file, "rt");

	if (f == NULL) return false;

	tMaterialInfo material;
	t3DObject     object;
	
	// mazani pameti
	memset(&material, 0, sizeof(material));
	memset(&object, 0, sizeof(object));

	// nacteni barvy trojuhelnucku
	fscanf(f, "%d %d %d", &material.color[0], &material.color[1], &material.color[2]);
   m_model.pMaterials.push_back(material);
	m_model.numOfMaterials = 1;

	// nacteni trojuhelnicku
   object.materialID = 0;      // je jen jeden matrial o ID 0
	object.bHasTexture = false; // netexturovany model
	object.numOfFaces = 1;      // 1 trojuhelnik
	object.numOfVerts = 3;      // 3 vrcholy
	object.numTexVertex = 0;    // zadne ukazatele na texturu
	
   object.pVerts = new CVector3[3];
   memset(object.pVerts, 0, sizeof(CVector3)*3);
	
	object.pFaces = new tFace [object.numOfFaces];
	memset(object.pFaces, 0, sizeof(tFace) * object.numOfFaces);

	// Go through all of the faces in this object
	for(int i = 0; i < 3; i++)	{
		fscanf(f, "%f %f %f", &object.pVerts[i].x, &object.pVerts[i].y, &object.pVerts[i].z);
		object.pFaces[0].vertIndex[i] = i;
	}

	m_model.numOfObjects = 1;
	m_model.pObject.push_back(object);

	// uklid
	fclose(f);

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////

bool H3DModel::Load3DObject3DS(const char *filename)
{
	memset(&m_model, 0, sizeof(m_model));

	CLoad3DS loader;
	loader.Import3DS(&m_model, filename);

	// Depending on how many textures we found, load each one
	// If you want to load other files than bitmaps, you will need to adjust CreateTexture().
	// Below, we go through all of the materials and check if they have a texture map to load.
	// Otherwise, the material just holds the color information and we don't need to load a texture.

	// Go through all the materials
	for(int i = 0; i < m_model.numOfMaterials; i++)
	{
		// Check to see if there is a file name to load in this material
		if(strlen(m_model.pMaterials[i].strFile) > 0)
		{
			// Use the name of the texture file to load the bitmap, with a texture ID (i).
			// We pass in our global texture array, the name of the texture, and an ID to reference it.
			m_model.pMaterials[i].textureId = htexture.GetTexture(m_model.pMaterials[i].strFile);
		}
		else m_model.pMaterials[i].textureId = -1;
	}
	
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////

void strlower (char *s)
{
	while (*s != '\0')
		*s = tolower(*s);
}

bool H3DModel::LoadModel(char* filename)
{

//	string s;
/*
	char c[255];
	strcpy(c, file);
	strup(c);

	tolower()
  */

//  strlower(filename);
  
	switch(filename[strlen(filename)-1]) {
  	case 's' : // *.3ds
   	if (!Load3DObject3DS(filename)) return false;
      m_LightEnabled = true;
    	break;
   case 't' :// *.txt
   	if (!Load3DObjectTXT(filename)) return false;
      m_LightEnabled = false;
    	break;   
	}
	// vypocitame normaly
	ComputeNormals();
	m_loaded = true;
	return true;
}

// *Note*
//
// Below are some math functions for calculating vertex normals.  We want vertex normals
// because it makes the lighting look really smooth and life like.  You probably already
// have these functions in the rest of your engine, so you can delete these and call
// your own.  I wanted to add them so I could show how to calculate vertex normals.


///////////////////////////////// COMPUTER NORMALS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function computes the normals and vertex normals of the objects
/////
///////////////////////////////// COMPUTER NORMALS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

void H3DModel::ComputeNormals()
{
	t3DModel *pModel = &m_model;

	CVector3 vVector1, vVector2, vNormal, vPoly[3];

	// If there are no objects, we can skip this part
	if(pModel->numOfObjects <= 0)
		return;

	// What are vertex normals?  And how are they different from other normals?
	// Well, if you find the normal to a triangle, you are finding a "Face Normal".
	// If you give OpenGL a face normal for lighting, it will make your object look
	// really flat and not very round.  If we find the normal for each vertex, it makes
	// the smooth lighting look.  This also covers up blocky looking objects and they appear
	// to have more polygons than they do.    Basically, what you do is first
	// calculate the face normals, then you take the average of all the normals around each
	// vertex.  It's just averaging.  That way you get a better approximation for that vertex.

	// Go through each of the objects to calculate their normals
	for(int index = 0; index < pModel->numOfObjects; index++)
	{
		// Get the current object
		t3DObject *pObject = &(pModel->pObject[index]);

		// Here we allocate all the memory we need to calculate the normals
		CVector3 *pNormals		= new CVector3 [pObject->numOfFaces];
		CVector3 *pTempNormals	= new CVector3 [pObject->numOfFaces];

		pObject->pNormals		= new CVector3 [pObject->numOfVerts];

		// Go though all of the faces of this object
		int i;
		for( i=0; i < pObject->numOfFaces; i++)
		{
			// To cut down LARGE code, we extract the 3 points of this face
			vPoly[0] = pObject->pVerts[pObject->pFaces[i].vertIndex[0]];
			vPoly[1] = pObject->pVerts[pObject->pFaces[i].vertIndex[1]];
			vPoly[2] = pObject->pVerts[pObject->pFaces[i].vertIndex[2]];

			// Now let's calculate the face normals (Get 2 vectors and find the cross product of those 2)

			vVector1 = Vector(vPoly[0], vPoly[2]);		// Get the vector of the polygon (we just need 2 sides for the normal)
			vVector2 = Vector(vPoly[2], vPoly[1]);		// Get a second vector of the polygon

			vNormal  = Cross(vVector1, vVector2);		// Return the cross product of the 2 vectors (normalize vector, but not a unit vector)
			pTempNormals[i] = vNormal;					// Save the un-normalized normal for the vertex normals
			vNormal  = Normalize(vNormal);				// Normalize the cross product to give us the polygons normal

			pNormals[i] = vNormal;						// Assign the normal to the list of normals
		}

		//////////////// Now Get The Vertex Normals /////////////////

		CVector3 vSum = {0.0, 0.0, 0.0};
		CVector3 vZero = vSum;
		int shared=0;

		for (i = 0; i < pObject->numOfVerts; i++)			// Go through all of the vertices
		{
			for (int j = 0; j < pObject->numOfFaces; j++)	// Go through all of the triangles
			{												// Check if the vertex is shared by another face
				if (pObject->pFaces[j].vertIndex[0] == i ||
					pObject->pFaces[j].vertIndex[1] == i ||
					pObject->pFaces[j].vertIndex[2] == i)
				{
					vSum = AddVector(vSum, pTempNormals[j]);// Add the un-normalized normal of the shared face
					shared++;								// Increase the number of shared triangles
				}
			}

			// Get the normal by dividing the sum by the shared.  We negate the shared so it has the normals pointing out.
			pObject->pNormals[i] = DivideVectorByScaler(vSum, float(-shared));

			// Normalize the normal for the final vertex normal
			pObject->pNormals[i] = Normalize(pObject->pNormals[i]);

			vSum = vZero;									// Reset the sum
			shared = 0;										// Reset the shared
		}

		// Free our memory and start over on the next object
		delete [] pTempNormals;
		delete [] pNormals;
	}
}
